# coding: utf-8

"""
Module for managing products and tiles in the library and temp folders.
"""

import os
import shutil
import logging
import pathlib
from itertools import chain

from pathlib import Path

# type annotations
from typing import List, Dict

from .config import Config, SHARED_DATA
from .products import L1cProduct
from .tiles import Tile
from .utils import human_size, human_size_decimal
from .multi_processing import (
    l2a_multiprocessing,
    cld_version_probability_iterations_reprocessing_multiprocessing,
    idx_multiprocessing,
)

s2_tiles_index = SHARED_DATA.get("tiles_index")

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class Library:
    """Class for listing L1C, L2A and indices tiles in the library folders."""

    _temp_path = Path(Config().get("temp_path"))
    _l1c_path = Path(Config().get("l1c_path"))
    _l2a_path = Path(Config().get("l2a_path"))
    _indices_path = Path(Config().get("indices_path"))

    def __init__(self):
        self._l1c = [f.name for f in self._l1c_path.glob("*")]
        self._l2a = [f.name for f in self._l2a_path.glob("*")]
        self._indices = dict()

        for indice in [f.name for f in self._indices_path.glob("*")]:
            self.__dict__[indice.lower()] = [
                f.name for f in self._indices_path.glob("{}/*".format(indice))
            ]

    @property
    def l1c(self) -> List[str]:
        """Returns tiles in the L1C library folder."""
        return self._l1c

    @property
    def l2a(self) -> List[str]:
        """Returns tiles in the L2A library folder."""
        return self._l2a

    @property
    def indices(self) -> Dict[str, pathlib.PosixPath]:
        """Returns tiles in the indices library folder."""
        return self._indices

    def clean(
        self,
        clean_list: list = [],
        remove_indice_tif: bool = False,
        remove: bool = False,
    ):
        """Function to clean corrupted files during processing errors from
        whole Library or selected Tiles. Calls the clean_lib function from Tile
        class.

        :param clean_list: list of tiles to be cleaned. If not provided, will
            process all l1c tiles.
        :param remove_indice_tif: bool. If True will remove present TIFF files
            present in indice folders. Default value False.
        :param remove: bool. If True will effectively remove corrupted files,
            if False will just list identified problems. Default value False.

        """
        nb_id = 0
        nb_rm = 0
        prob_id = []
        if not clean_list:
            clean_list = self.l1c
        for t in clean_list:
            try:
                til = Tile(t)
                counts = til.clean_lib(
                    remove=remove, remove_indice_tif=remove_indice_tif
                )
                nb_id += counts["identified_problems"]
                nb_rm += counts["removed_problems"]
                if counts["problems"]:
                    prob_id.append(counts["problems"])
                logger.info(
                    "{} - identified problems:{} - removed problems:{}".format(
                        til,
                        counts["identified_problems"],
                        counts["removed_problems"],
                    )
                )
            except:
                pass
        return {
            "identified_problems": nb_id,
            "removed_problems": nb_rm,
            "problems": prob_id,
        }

    def archive_l1c(
        self,
        archive_list: list = [],
        force: bool = False,
        size_only: bool = False,
    ):
        """
        Function to archive l1c products from library folder to l1c_archive_path.
        Calls the archive_l1c function from Tile class, see there for details.

        :param archive_list: list of tiles to archive. If not provided, will
            process all l1c tiles.
        :param size_only: if True, only gives sizes
        """

        total_size = 0
        if not archive_list:
            archive_list = self.l1c
        for t in archive_list:
            try:
                logger.info(t)
                til = Tile(t)
                size_tile = til.archive_l1c(size_only=size_only, force=force)
                if size_only:
                    total_size += size_tile
            except:
                pass
        logger.info(
            "Total l1c size to move: {}".format(human_size_decimal(total_size))
        )

    def archive_l2a(
        self,
        tile_list: list = [],
        size_only: bool = False,
    ):

        """
        Function to archive l2a products from library folder to l2a_archive_path.
        Calls the archive_l2a function from Tile class, see there for details.

        :param tile_list: list of tiles to archive. If not provided, will
            process all l2a tiles.
        :param size_only: if True, only gives sizes.
        """

        total_size = 0
        if not tile_list:
            tile_list = self.l2a
        for t in tile_list:
            try:
                logger.info(t)
                til = Tile(t)
                size_tile = til.archive_l2a(size_only=size_only)
                if size_only:
                    total_size += size_tile
            except:
                logger.info("Problem while archiving")
        logger.info(
            "Total l2a size to move: {}".format(human_size_decimal(total_size))
        )

    def archive_all(
        self,
        archive_list: list = [],
        size_only: bool = False,
        force: bool = False,
    ):
        total_size = 0
        if not archive_list:
            archive_list = self.l1c
        for t in archive_list:
            try:
                logger.info(t)
                til = Tile(t)
                size_tile = til.archive_all(size_only=size_only, force=force)
                total_size += size_tile
            except:
                pass
        logger.info(
            "Total size to move: {}".format(human_size_decimal(total_size))
        )

    def compute_ql(
        self,
        tile_list: list = [],
        product_list: list = [],
        resolution: int = 750,
        jpg: bool = True,
    ):
        """
        Produce:
         - l1c and/or l2a quicklooks
         - for provided tile_list or all the library tiles
         - at specified resolution
         - in jpg or tiff format
        """
        if not tile_list:
            tile_list = list(set(self.l1c + self.l2a))
        for tile in tile_list:
            Tile(tile).compute_ql(product_list, resolution, jpg)

    def update_latest_ql(self):
        """
        Produce or update the latest quicklook for the L2A library tiles
        """
        try:
            for tile in self.l2a:
                Tile(tile).update_latest_ql()
        except:
            pass

    # def update_latest_ql(self):
    #     """
    #     Produce or update the latest quicklook for the L2A library tiles
    #     """
    #     for tile in self.l2a:
    #         Tile(tile).update_latest_ql()

    def update_old_cloudmasks(self):
        """
        Produce or update the latest quicklook for the L2A library tiles
        """
        for tile in self.l2a:
            Tile(tile).update_old_cloudmasks()

    def remove_very_old_cloudmasks(self):
        """
        Remove very old cloudmasks, matching pattern : *_CLOUD_MASK.tif for the L2A library tiles
        """
        for tile in self.l2a:
            Tile(tile).remove_very_old_cloudmasks()

    def move_old_quicklooks(self):
        """
        Move all old quicklooks to QL subfolder for the L2A library tiles
        """
        for tile in self.l2a:
            Tile(tile).move_old_quicklooks()

    def update_old_indices(self):
        """
        Rename old indices to match new cloudmask nomenclature
        for the L2A library tiles
        """
        for tile in {f.name for f in list(self._indices_path.glob("*/*/"))}:
            Tile(tile).update_old_indices()

    def init_md(self):
        """
        Initiate sen2chain metadata for all tile products (l2a, cloudmasks, indices (raw, masked, ql))
        for the L1C, L2A and Indice library tiles
        """
        for tile in set(
            [
                val
                for sublist in [
                    k for k in [getattr(self, t) for t in self.__dict__]
                ]
                for val in sublist
            ]
        ):
            Tile(tile).init_md()

    def compute_l2a(
        self,
        tile_list: list = [],
        date_min: str = None,
        date_max: str = None,
        nb_proc: int = 4,
    ):
        l1c_process_list = []
        for tile in tile_list:
            t = Tile(tile)
            l1c_process_list.append(
                list(
                    p.identifier
                    for p in t.l2a_missings.filter_dates(
                        date_min=date_min, date_max=date_max
                    )
                )
            )
        l1c_process_list = list(chain.from_iterable(l1c_process_list))
        logger.info(
            "l1c_process_list ({} files): \n{}".format(
                len(l1c_process_list), l1c_process_list
            )
        )

        if l1c_process_list:
            l2a_res = l2a_multiprocessing(l1c_process_list, nb_proc=nb_proc)

    def compute_cloudmasks(
        self,
        tile_list: list = [],
        cm_version: str = "cm001",
        probability: int = 1,
        iterations: int = 5,
        cld_shad: bool = True,
        cld_med_prob: bool = True,
        cld_hi_prob: bool = True,
        thin_cir: bool = True,
        reprocess: bool = False,
        date_min: str = None,
        date_max: str = None,
        nb_proc: int = 4,
    ):
        cld_l2a_process_list = []
        for tile in tile_list:
            t = Tile(tile)
            if not reprocess:
                l2a_list = [
                    p.identifier
                    for p in t.cloudmasks_missing(
                        cm_version=cm_version,
                        probability=probability,
                        iterations=iterations,
                        cld_shad=cld_shad,
                        cld_med_prob=cld_med_prob,
                        cld_hi_prob=cld_hi_prob,
                        thin_cir=thin_cir,
                    ).filter_dates(date_min=date_min, date_max=date_max)
                ]
            else:
                l2a_list = [
                    p.identifier
                    for p in t.l2a.filter_dates(
                        date_min=date_min, date_max=date_max
                    )
                ]
            for j in l2a_list:
                cld_l2a_process_list.append(
                    [
                        j,
                        cm_version,
                        probability,
                        iterations,
                        cld_shad,
                        cld_med_prob,
                        cld_hi_prob,
                        thin_cir,
                        reprocess,
                    ]
                )
        logger.info(
            "cld_l2a_process_list ({} files): \n{}".format(
                len(cld_l2a_process_list), cld_l2a_process_list
            )
        )
        if cld_l2a_process_list:
            cld_res = cld_version_probability_iterations_reprocessing_multiprocessing(
                cld_l2a_process_list, nb_proc=nb_proc
            )

    def compute_indices(
        self,
        tile_list: list = [],
        indice_list: list = [],
        reprocess: bool = False,
        nodata_clouds: bool = True,
        quicklook: bool = False,
        cm_version: list = "cm001",
        probability: int = 1,
        iterations: int = 5,
        cld_shad: bool = True,
        cld_med_prob: bool = True,
        cld_hi_prob: bool = True,
        thin_cir: bool = True,
        date_min: str = None,
        date_max: str = None,
        nb_proc: int = 4,
    ):
        indices_l2a_process_list = []
        for tile in tile_list:
            t = Tile(tile)
            for i in indice_list:
                if not reprocess:
                    l2a_list = [
                        p.identifier
                        for p in t.missing_indices(
                            i,
                            nodata_clouds=nodata_clouds,
                            cm_version=cm_version,
                            probability=probability,
                            iterations=iterations,
                            cld_shad=cld_shad,
                            cld_med_prob=cld_med_prob,
                            cld_hi_prob=cld_hi_prob,
                            thin_cir=thin_cir,
                        ).filter_dates(date_min=date_min, date_max=date_max)
                    ]
                else:
                    l2a_list = [
                        p.identifier
                        for p in t.l2a.filter_dates(
                            date_min=date_min, date_max=date_max
                        )
                    ]

                for j in l2a_list:
                    indices_l2a_process_list.append(
                        [
                            j,
                            i,
                            reprocess,
                            nodata_clouds,
                            quicklook,
                            cm_version,
                            probability,
                            iterations,
                            cld_shad,
                            cld_med_prob,
                            cld_hi_prob,
                            thin_cir,
                        ]
                    )
        logger.info(
            "indices_l2a_process_list ({} files): \n{}".format(
                len(indices_l2a_process_list), indices_l2a_process_list
            )
        )
        if indices_l2a_process_list:
            indices_res = idx_multiprocessing(
                indices_l2a_process_list, nb_proc=nb_proc
            )


class TempContainer:
    """Class for managing a downloaded L1C products.

    :param l1c_identifier: L1C product's identifier.
    :param tile: product's tile name.
    """

    _temp_path = Path(Config().get("temp_path"))

    def __init__(self, l1c_identifier=None, tile: str = None):
        self.tile = tile
        self.temp_id = l1c_identifier + "_TEMP"
        self.temp_path = self._temp_path / self.temp_id
        self.l1c = L1cProduct(
            identifier=l1c_identifier, tile=tile, path=self.temp_path
        )

    def create_temp_folder(self) -> "TempContainer":
        """Creates the L1C product's temporary folder in the TEMP library folder."""
        logger.info("{}: creating temp folder".format(self.l1c.identifier))
        try:
            self.temp_path.mkdir(exist_ok=True)
        except FileNotFoundError:
            print("Invalid temp folder path:", self.temp_path)
            raise
        return self

    def unzip_l1c(self) -> "TempContainer":
        """Unzips the downloaded L1C zip file in the product's temporary folder."""
        zip_file = self.temp_path / (self.l1c.identifier + ".zip")
        if zip_file.exists():
            logger.info(
                "{} : Unzipping L1C archive".format(self.l1c.identifier)
            )
            shutil.unpack_archive(str(zip_file), str(self.temp_path))
            sorted(zip_file.parent.glob("*.SAFE"))[0].rename(
                zip_file.parent / (zip_file.stem + ".SAFE")
            )
            os.remove(str(zip_file))
        return self

    def delete_temp_folder(self) -> "TempContainer":
        """Deletes the L1C product's temporary folder."""
        logger.info("Deleting temp folder: {}".format(self.temp_path))
        if not list(self.temp_path.glob("*.SAFE")):
            try:
                shutil.rmtree(str(self.temp_path))
            except FileNotFoundError:
                logger.warning(
                    "Temp folder does not exist: {}".format(self.temp_path)
                )
        else:
            logger.error("Temp folder contains a SAFE product.")
        return self
