# coding: utf-8

"""
Module for managing products and tiles in the library and temp folders.
"""

import pathlib
import shutil
import logging
import re
import os

from pathlib import Path

# type annotations
from typing import (
    List, 
    Tuple, 
    Optional, 
    Union
)
from packaging import version
from .utils import grouper, set_permissions, get_Sen2Cor_version, get_latest_s2c_version_path
from .config import Config, SHARED_DATA
from .xmlparser import MetadataParser, Sen2ChainMetadataParser
from .sen2cor import process_sen2cor
from .cloud_mask import (
    create_cloud_mask,
    create_cloud_mask_v2,
    create_cloud_mask_b11,
    create_cloud_mask_v003,
    create_cloud_mask_v004,
)
from .indices import IndicesCollection
from .colormap import (
    create_l2a_ql,
    create_l1c_ql,
    create_l1c_ql_v2,
    create_l2a_ql_v2,
)

s2_tiles_index = SHARED_DATA.get("tiles_index")

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class Product:
    """Base product template class. This class is designed to be
    instanciated only by sub-classes.

    :param identifier: L1C product's identifier or SAFE filename.
    :param tile: tile's name.
    :param path: parent's SAFE folder path.
    """

    _library_path = Path("")
    _metadata_filename_templates = ()

    def __new__(cls, *args, **kwargs):
        """Prevents from direct instantiation."""
        if cls is Product:
            raise TypeError("Product may not be instantiated")
        return object.__new__(cls)

    def __init__(
        self,
        identifier: str = None,
        tile: str = None,
        path: pathlib.PosixPath = None,
    ) -> None:
        if identifier is None:
            raise ValueError("Product identifier is empty")
        else:
            self.identifier = identifier.upper().replace(".SAFE", "")

        if tile is None:
            try:
                self.tile = Product.get_tile(identifier)
                self._tiled = True
            except IndexError:
                logger.error("Tile must be specified for old products")
                raise
        else:
            self.tile = tile
            try:
                Product.get_tile(identifier)
                self._tiled = True
            except IndexError:
                self._tiled = False

        self.safe = self.identifier + ".SAFE"
        self.library_path = self._library_path / self.tile / self.safe

        if path is None:
            self.path = self.library_path
            logger.debug(
                "No path specified, using library's: {}".format(self.path)
            )
        else:
            self.path = path / self.safe

        if self.path.absolute() is self.library_path.absolute():
            self._library_product = True
        else:
            self._library_product = False

        self._metadata_path = None
        self._metadata_parser = None

    @staticmethod
    def get_tile(string) -> str:
        """Returns tile name from a string.

        :param string: string from which to extract the tile name.
        """
        return re.findall("_T([0-9]{2}[A-Z]{3})_", string)[0]

    def _update_metadata_parser(self) -> None:
        """Instanciates the metadaparser object."""
        if self._metadata_parser is None:
            try:
                self._get_metadata_path()
                self._metadata_parser = MetadataParser(
                    metadata_path=self._metadata_path, tile=self.tile
                )
            except Exception as e:
                logger.debug("{}".format(e))
                logger.error(
                    "{}: could not load metadata: {}".format(
                        self.identifier, self._metadata_path
                    )
                )

    def _get_metadata_path(self) -> None:
        """Searchs and sets the metadata path of the product."""
        logger.debug("{}: looking for metadata path".format(self.identifier))
        if self._tiled:
            filename = self._tiled_metadata
        else:
            filename = (
                self.identifier.replace("_PRD_", "_MTD_").replace(
                    "_MSI", "_SAF"
                )
                + ".xml"
            )
        metadata_path = self.path / filename
        if not metadata_path.exists():
            self._metadata_path = None
            logger.debug("{} metadata file not found".format(self.safe))
        self._metadata_path = metadata_path
        logger.debug("Found: {}".format(self._metadata_path))

    def _get_metadata_value(self, key: str) -> str:
        """Returns a metadata value.

        :param key: tag of the metadata to search.
        """
        self._update_metadata_parser()
        return self._metadata_parser.get_metadata_value(key=key)

    def _get_band_path(self, key: str, res=None) -> str:
        """Returns a band path according to a key.

        :param key: tag of the metadata to search.
        """
        self._update_metadata_parser()
        try:
            return self._metadata_parser.get_band_path(key=key, res=res)
        except:
            return None

    @property
    def in_library(self) -> bool:
        """Is the product in the library folder."""
        return self.library_path.exists()

    def archive(self) -> "Product":
        """Moves the product's SAFE folder to the library folder."""
        logger.info("archiving {} to {}".format(self.path, self.library_path))
        self.library_path.parent.mkdir(exist_ok=True)
        shutil.move(str(self.path), str(self.library_path.parent))
        return self

    def __str__(self) -> Optional[str]:
        try:
            return str(self.safe)
        except TypeError:
            return None


class L1cProduct(Product):
    """L1C product class.

    :param identifier: L1C product's identifier or SAFE filename.
    :param tile: product's tile.
    :param path: parent folder.
    """

    _library_path = Path(Config().get("l1c_path"))
    _tiled_metadata = "MTD_MSIL1C.xml"

    def __init__(
        self, identifier: str = None, tile: str = None, path: str = None
    ):
    
        super().__init__(identifier=identifier, tile=tile, path=path)
        
        if not re.match(r".*L1C_.*", self.identifier):
            raise ValueError("Invalid L1C product name")
    
    def processable_to_l2a(
        self, 
        s2c_path: Union[str, pathlib.PosixPath] = None,
    ) -> "L1cProduct":
        """ check if l1c can be processed to l2a with sen2cor """
        s2c_path = s2c_path or get_latest_s2c_version_path(self.identifier)
        if s2c_path:
            return True
        else:
            logger.info("{} not processable, check your Sen2Cor install".format(self.identifier))
            return False
    
    def process_l2a(
        self, 
        reprocess: bool = False,
        s2c_path: Union[str, pathlib.PosixPath] = None,
    ) -> "L1cProduct":
        """ process with sen2cor """
        logger.info("{}: processing L2A".format(self.identifier))

        l2a_identifier = self.identifier.replace("L1C_", "L2A_").replace(
            "_OPER_", "_USER_"
        )
        l2a_path = self.path.parent / (l2a_identifier + ".SAFE")
        l2a_prod = L2aProduct(l2a_identifier, self.tile, l2a_path.parent)
        
        process_it = False
        if not l2a_prod.in_library and not l2a_path.exists():
            s2c_path = s2c_path or get_latest_s2c_version_path(self.identifier)
            if s2c_path:
                process_it = True
            else:
                logger.info("{} not processable, check your Sen2Cor install".format(self.identifier))
        else:
            if not reprocess:
                logger.info("{} already exists.".format(l2a_identifier))
            else:
                if l2a_prod.in_library:
                    shutil.rmtree(str(l2a_prod.path))
                    logger.info("Deleted {}".format(l2a_prod.path))
                else:
                    # if reprocessing from the temp folder
                    shutil.rmtree(str(l2a_path))
                    logger.info("Deleted {}".format(l2a_path))
                
                #### check here if better to delete after sen2cor version check compatible... ??
                s2c_path = s2c_path or get_latest_s2c_version_path(self.identifier)
                if s2c_path:
                    process_it = True

        if process_it:
            process_sen2cor(
                l1c_product_path = self.path, 
                l2a_product_path = l2a_path, 
                s2c_path = s2c_path,
                pb = self.processing_baseline,
            )
            # sen2cor creates the L2A in the parent folder of the L1C.
            # Therefore, if the L1C computed is in the L1C library folder,
            # the newly L2A must be moved to his L2A library folder.
            if self._library_product:
                l2a_prod.archive()
            l2a_prod = L2aProduct(l2a_identifier)
            l2a_prod.set_permissions()
            l2a_prod.update_md(sen2cor_version = get_Sen2Cor_version(s2c_path))
        return self

    def process_ql(
        self,
        reprocess: bool = False,
        out_path: pathlib.PosixPath = None,
        out_resolution: Tuple[int, int] = (100, 100),
        jpg=False,
    ) -> "L1cProduct":
        """ """
        logger.info("{}: processing L1C Quicklook".format(self.identifier))

        if jpg:
            ql_filename = (
                self.identifier + "_QL-" + str(out_resolution[0]) + "m.jpg"
            )
        else:
            ql_filename = (
                self.identifier + "_QL-" + str(out_resolution[0]) + "m.tif"
            )

        if out_path is None:
            ql_folder = self.library_path.parent / "QL"
            ql_folder.mkdir(parents=True, exist_ok=True)
            ql_path = ql_folder / ql_filename
        else:
            ql_path = Path(out_path)

        if ql_path.exists() and not reprocess:
            logger.info("{} Quicklook already exists".format(self.identifier))
        else:
            # create_l1c_ql(tci = self.tci,
            # out_path = ql_path,
            # out_resolution = out_resolution,
            # jpg = jpg)
            create_l1c_ql_v2(
                tci=self.tci,
                out_path=ql_path,
                out_resolution=out_resolution,
                jpg=jpg,
            )
            self.user_ql = ql_path

        return ql_path

    def remove(self):
        if self.path.is_symlink():
            l1c_path = os.readlink(str(self.path))
            logger.info("Removing: {}".format(l1c_path))
            shutil.rmtree(l1c_path)
            logger.info("Removing symlink: {}".format(self.path))
            self.path.unlink()
        else:
            # l1c_path = os.readlink(str(l1c.path))
            logger.info("Removing: {}".format(self.path))
            shutil.rmtree(str(self.path))

    # METADATA
    @property
    def product_start_time(self):
        return self._get_metadata_value(key="PRODUCT_START_TIME")

    @property
    def product_stop_time(self):
        return self._get_metadata_value(key="PRODUCT_STOP_TIME")

    @property
    def product_uri(self):
        return self._get_metadata_value(key="PRODUCT_URI")

    @property
    def product_type(self):
        return self._get_metadata_value(key="PRODUCT_TYPE")

    @property
    def processing_baseline(self):
        return self._get_metadata_value(key="PROCESSING_BASELINE")

    @property
    def generation_time(self):
        return self._get_metadata_value(key="GENERATION_TIME")

    @property
    def spacecraft_name(self):
        return self._get_metadata_value(key="SPACECRAFT_NAME")

    @property
    def datatake_type(self):
        return self._get_metadata_value(key="DATATAKE_TYPE")

    @property
    def datatake_sensing_start(self):
        return self._get_metadata_value(key="DATATAKE_SENSING_START")

    @property
    def sensing_orbit_number(self):
        return self._get_metadata_value(key="SENSING_ORBIT_NUMBER")

    @property
    def sensing_orbit_direction(self):
        return self._get_metadata_value(key="SENSING_ORBIT_DIRECTION")

    @property
    def cloud_coverage_assessment(self):
        return self._get_metadata_value(key="Cloud_Coverage_Assessment")

    @property
    def footprint(self):
        return self._get_metadata_value(key="EXT_POS_LIST")

    # BANDS PATHS
    @property
    def b01(self):
        return self._get_band_path(key="B01")

    @property
    def b02(self):
        return self._get_band_path(key="B02")

    @property
    def b03(self):
        return self._get_band_path(key="B03")

    @property
    def b04(self):
        return self._get_band_path(key="B04")

    @property
    def b05(self):
        return self._get_band_path(key="B05")

    @property
    def b06(self):
        return self._get_band_path(key="B06")

    @property
    def b07(self):
        return self._get_band_path(key="B07")

    @property
    def b08(self):
        return self._get_band_path(key="B08")

    @property
    def b8a(self):
        return self._get_band_path(key="B8A")

    @property
    def b09(self):
        return self._get_band_path(key="B09")

    @property
    def b10(self):
        return self._get_band_path(key="B10")

    @property
    def b11(self):
        return self._get_band_path(key="B11")

    @property
    def b12(self):
        return self._get_band_path(key="B12")

    @property
    def tci(self):
        return self._get_band_path(key="TCI")


class L2aProduct(Product):
    """L2A product class.

    :param identifier: L2A product's identifier or SAFE filename.
    :param tile: tile's name.
    :param path: parent's SAFE folder path.
    """

    _library_path = Path(Config().get("l2a_path"))
    _indices_library_path = Path(Config().get("indices_path"))
    _tiled_metadata = "MTD_MSIL2A.xml"

    def __init__(
        self, identifier: str = None, tile: str = None, path: str = None
    ) -> None:
        super().__init__(identifier=identifier, tile=tile, path=path)
        if not re.match(r".*L2A_.*", identifier):
            raise ValueError("Invalid L2A product name")

        self.indices_path = self._indices_library_path

        # user cloud mask
        # self.user_cloud_mask = self.path.parent / (self.identifier + "_CLOUD_MASK.jp2")
        # if not self.user_cloud_mask.exists():
        # self.user_cloud_mask = None
        # # user cloud mask b11
        # self.user_cloud_mask_b11 = self.path.parent / (self.identifier + "_CLOUD_MASK_B11.jp2")
        # if not self.user_cloud_mask_b11.exists():
        # self.user_cloud_mask_b11 = None
        # user QL
        self.user_ql = self.path.parent / (self.identifier + "_QL.tif")
        if not self.user_ql.exists():
            self.user_ql = None

        # versions
        self._sen2chain_info_path = self.path / "sen2chain_info.xml"
        if (
            self._sen2chain_info_path.parent.exists()
            and not self._sen2chain_info_path.exists()
        ):
            Sen2ChainMetadataParser(self._sen2chain_info_path).init_metadata()

    def process_ql(
        self,
        reprocess: bool = False,
        out_path: pathlib.PosixPath = None,
        out_resolution: Tuple[int, int] = (100, 100),
        jpg=False,
    ) -> "L2aProduct":
        """ """
        logger.info("{}: processing L2A Quicklook".format(self.identifier))

        if jpg:
            ql_filename = (
                self.identifier + "_QL-" + str(out_resolution[0]) + "m.jpg"
            )
        else:
            ql_filename = (
                self.identifier + "_QL-" + str(out_resolution[0]) + "m.tif"
            )

        if out_path is None:
            ql_folder = self.library_path.parent / "QL"
            ql_folder.mkdir(parents=True, exist_ok=True)
            ql_path = ql_folder / ql_filename
        else:
            ql_path = Path(out_path)

        if ql_path.exists() and not reprocess:
            logger.info("{} Quicklook already exists".format(self.identifier))
        else:
            # create_l2a_ql(b02 = self.b02_10m,
            # b03 = self.b03_10m,
            # b04 = self.b04_10m,
            # out_path = ql_path,
            # out_resolution = out_resolution,
            # jpg = jpg)

            create_l2a_ql_v2(
                tci=self.tci_10m,
                out_path=ql_path,
                out_resolution=out_resolution,
                jpg=jpg,
            )

            self.user_ql = ql_path

        return self

    def process_cloud_mask(
        self,
        buffering: bool = True,
        reprocess: bool = False,
        out_path: pathlib.PosixPath = None,
    ) -> "L2aProduct":
        """ """
        logger.info("{}: processing cloud_mask".format(self.identifier))

        cloud_mask_filename = self.identifier + "_CLOUD_MASK.tif"

        if out_path is None:
            cloud_mask_folder = self.library_path.parent
            cloud_mask_folder.mkdir(parents=True, exist_ok=True)
            cloud_mask_path = cloud_mask_folder / cloud_mask_filename
        else:
            cloud_mask_path = Path(out_path)

        if cloud_mask_path.exists() and not reprocess:
            logger.info("{} cloud mask already exists".format(self.identifier))
        else:
            create_cloud_mask(
                self.cld_20m,
                buffering=buffering,
                erosion=-40,
                dilatation=100,
                out_path=cloud_mask_path,
            )
            self.user_cloud_mask = cloud_mask_path
        return self

    def process_cloud_mask_v2(
        self,
        buffering: bool = True,
        reprocess: bool = False,
        out_path_mask=None,
        out_path_mask_b11=None,
    ) -> "L2aProduct":
        """ """
        logger.info("{}: processing cloud_mask_v2".format(self.identifier))

        cloud_mask_filename = self.identifier + "_CLOUD_MASK.jp2"
        cloud_mask_b11_filename = (
            str(
                Path(cloud_mask_filename).parent
                / Path(cloud_mask_filename).stem
            )
            + "_B11.jp2"
        )

        if out_path_mask is None:
            cloud_mask_folder = self.library_path.parent
            cloud_mask_folder.mkdir(parents=True, exist_ok=True)
            cloud_mask_path = cloud_mask_folder / cloud_mask_filename
        else:
            cloud_mask_path = Path(out_path_mask)

        if out_path_mask_b11 is None:
            cloud_mask_folder = self.library_path.parent
            cloud_mask_folder.mkdir(parents=True, exist_ok=True)
            cloud_mask_b11_path = cloud_mask_folder / cloud_mask_b11_filename
        else:
            cloud_mask_b11_path = Path(out_path_mask_b11)

        if cloud_mask_path.exists() and not reprocess:
            logger.info(
                "{} cloud mask v2 already exists".format(self.identifier)
            )
        else:
            create_cloud_mask_v2(
                self.cld_20m, erosion=1, dilatation=5, out_path=cloud_mask_path
            )
            self.user_cloud_mask = cloud_mask_path

        if cloud_mask_b11_path.exists() and not reprocess:
            logger.info(
                "{} cloud mask v2 masked b11 already exists".format(
                    self.identifier
                )
            )
        else:
            create_cloud_mask_b11(
                self.user_cloud_mask,
                self.b11_20m,
                dilatation=5,
                out_path=cloud_mask_b11_path,
            )
            self.user_cloud_mask_b11 = cloud_mask_b11_path

        return self

    def compute_cloud_mask(
        self,
        cm_version: str = "cm001",
        probability: int = 1,
        iterations: int = 5,
        cld_shad: bool = True,
        cld_med_prob: bool = True,
        cld_hi_prob: bool = True,
        thin_cir: bool = True,
        reprocess: bool = False,
        # out_path_mask = None,
        # out_path_mask_b11 = None
    ) -> "L2aProduct":
        """ """
        if cm_version == "cm003":
            logger.info(
                "Computing cloudmask version {}, probability {}%, iteration(s) {}: {}".format(
                    cm_version, probability, iterations, self.identifier
                )
            )
        elif cm_version == "cm004":
            logger.info(
                "Computing cloudmask version {}, ITER {}, SHAD {}, MED-PRB {}, HI-PRB {}, CIRR {}: {}".format(
                    cm_version,
                    iterations,
                    cld_shad,
                    cld_med_prob,
                    cld_hi_prob,
                    thin_cir,
                    self.identifier,
                )
            )
        else:
            logger.info(
                "Computing cloudmask version {}: {}".format(
                    cm_version, self.identifier
                )
            )

        cloudmask = NewCloudMaskProduct(
            l2a_identifier=self.identifier,
            sen2chain_processing_version=self.sen2chain_processing_version,
            cm_version=cm_version,
            probability=probability,
            iterations=iterations,
            cld_shad=cld_shad,
            cld_med_prob=cld_med_prob,
            cld_hi_prob=cld_hi_prob,
            thin_cir=thin_cir,
        )

        if cloudmask.path.exists() and not reprocess:
            logger.info(
                "{} cloud mask already computed".format(cloudmask.identifier)
            )
        else:
            if not cloudmask.path.parent.exists():
                cloudmask.path.parent.mkdir(parents=True)
            if cm_version == "cm001":
                if (
                    cloudmask.path.exists()
                ):  # in version 3.8 will be updated using missing_ok = True
                    cloudmask.path.unlink()
                    cloudmask._info_path.unlink()
                create_cloud_mask_v2(
                    self.cld_20m,
                    erosion=1,
                    dilatation=5,
                    out_path=cloudmask.path,
                )
                # self.user_cloud_mask = cloudmask.updated_path
            elif cm_version == "cm002":
                cloudmask_cm001 = NewCloudMaskProduct(
                    l2a_identifier=self.identifier,
                    sen2chain_processing_version=self.sen2chain_processing_version,
                    cm_version="cm001",
                )
                if cloudmask_cm001.path.exists():
                    if (
                        cloudmask.path.exists()
                    ):  # in version 3.8 will be updated using missing_ok = True
                        cloudmask.path.unlink()
                        cloudmask._info_path.unlink()
                    create_cloud_mask_b11(
                        cloudmask_cm001.path,
                        self.b11_20m,
                        dilatation=5,
                        out_path=Path(
                            str(cloudmask_cm001.path).replace(
                                "CM001", "CM002-B11"
                            )
                        ),
                    )
                else:
                    logger.info(
                        "No cloudmask version cm001 found, please compute this one first"
                    )
            elif cm_version == "cm003":
                if (
                    cloudmask.path.exists()
                ):  # in version 3.8 will be updated using missing_ok = True
                    cloudmask.path.unlink()
                    cloudmask._info_path.unlink()
                create_cloud_mask_v003(
                    cloud_mask=self.msk_cldprb_20m,
                    out_path=cloudmask.path,
                    probability=probability,
                    iterations=iterations,
                )

            elif cm_version == "cm004":
                if (
                    cloudmask.path.exists()
                ):  # in version 3.8 will be updated using missing_ok = True
                    cloudmask.path.unlink()
                    cloudmask._info_path.unlink()
                create_cloud_mask_v004(
                    scl_path=self.scl_20m,
                    out_path=cloudmask.path,
                    iterations=iterations,
                    cld_shad=cld_shad,
                    cld_med_prob=cld_med_prob,
                    cld_hi_prob=cld_hi_prob,
                    thin_cir=thin_cir,
                )
            else:
                logger.info("Wrong cloudmask version {}".format(cm_version))

            cloudmask.init_md()

        # return self

    def process_indices(
        self,
        indices_list: List[str] = [],
        nodata_clouds: bool = False,
        quicklook: bool = False,
        reprocess: bool = False,
        out_path: str = None,
    ) -> "L2aProduct":
        """
        Process indices.

        :param indices_list: list of valid indices names.
        :param nodata_clouds: mask indices rasters with the cloud-mask or not.
        :param quicklook: process a quicklook or not (if available for the indice).
        :param reprocess: erase old data.
        :param out_path: if specified, a folder {identifier}_INDICES containing \
        the processed indices will be created in the out_path.
        """
        logger.info("{}: processing_indices".format(self.identifier))

        if not isinstance(indices_list, list):
            raise TypeError("Indices must be provided as a list.")

        for indice in set(indices_list):
            logger.info("Processing {}: {}".format(indice, self.identifier))
            indice_cls = IndicesCollection.get_indice_cls(indice)
            if indice_cls is None:
                print(
                    "Indices available: {}".format(
                        IndicesCollection.available_indices
                    )
                )
                raise ValueError("Indice not defined")

            if out_path is None:
                # logger.info(self.identifier)
                indice_path = (
                    self.indices_path
                    / indice.upper()
                    / self.tile
                    / self.identifier
                )
            else:
                indice_path = (
                    Path(out_path)
                    / (self.identifier + "_INDICES")
                    / indice.upper()
                )
            indice_path.mkdir(parents=True, exist_ok=True)
            indice_obj = indice_cls(self)
            indice_obj.process_indice(
                out_path=indice_path,
                nodata_clouds=nodata_clouds,
                quicklook=quicklook,
                reprocess=reprocess,
            )
        return self

    def compute_indice(
        self,
        indice: str = None,
        reprocess: bool = False,
        nodata_clouds: bool = True,
        quicklook: bool = False,
        cm_version: str = "cm001",
        probability: int = 1,
        iterations: int = 5,
        cld_shad: bool = True,
        cld_med_prob: bool = True,
        cld_hi_prob: bool = True,
        thin_cir: bool = True,
        out_path: str = None,
    ) -> "L2aProduct":
        """
        compute and mask indice specified cloudmask version
        :param indice: a valid indice name
        :param nodata_clouds: mask indices rasters with the cloud-mask or not
        :param cm_version: cloudmask version to use for masking
        :param quicklook: process a quicklook or not (if available for the indice).
        :param reprocess: if true reprocess
        :param out_path: if specified, a folder {identifier}_INDICES containing \
        the processed indices will be created in the out_path.
        """
        logger.info("Computing indice {}: {}".format(indice, self.identifier))

        # if not isinstance(indices_list, list):
        # raise TypeError("Indices must be provided as a list.")

        # for indice in set(indices_list):
        # logger.info("Processing {}: {}".format(indice, self.identifier))
        indice_cls = IndicesCollection.get_indice_cls(indice)
        if indice_cls is None:
            print(
                "Indices available: {}".format(
                    IndicesCollection.available_indices
                )
            )
            raise ValueError("Indice not defined")

        if out_path is None:
            # logger.info(self.identifier)
            indice_path = (
                self.indices_path
                / indice.upper()
                / self.tile
                / self.identifier
            )
        else:
            indice_path = (
                Path(out_path)
                / (self.identifier + "_INDICES")
                / indice.upper()
            )
        indice_path.mkdir(parents=True, exist_ok=True)

        indice_obj = indice_cls(
            self,
            NewCloudMaskProduct(
                l2a_identifier=self.identifier,
                cm_version=cm_version,
                probability=probability,
                iterations=iterations,
                cld_shad=cld_shad,
                cld_med_prob=cld_med_prob,
                cld_hi_prob=cld_hi_prob,
                thin_cir=thin_cir,
            ),
        )

        indice_obj.process_indice(
            out_path=indice_path,
            nodata_clouds=nodata_clouds,
            quicklook=quicklook,
            reprocess=reprocess,
        )
        if nodata_clouds:
            indice_raw = IndiceProduct(
                l2a_identifier=self.identifier,
                indice=indice,
                masked=not nodata_clouds,
                cm_version=cm_version,
                probability=probability,
                iterations=iterations,
                cld_shad=cld_shad,
                cld_med_prob=cld_med_prob,
                cld_hi_prob=cld_hi_prob,
                thin_cir=thin_cir,
            )

            indice_masked = IndiceProduct(
                l2a_identifier=self.identifier,
                indice=indice,
                masked=nodata_clouds,
                cm_version=cm_version,
                probability=probability,
                iterations=iterations,
                cld_shad=cld_shad,
                cld_med_prob=cld_med_prob,
                cld_hi_prob=cld_hi_prob,
                thin_cir=thin_cir,
            )
            if quicklook:
                indice_masked_ql = IndiceProduct(
                    identifier=indice_masked.identifier.replace(
                        ".jp2", "_QL.tif"
                    )
                )

        else:
            IndiceProduct(
                l2a_identifier=self.identifier,
                indice=indice,
                masked=nodata_clouds,
                cm_version=cm_version,
                probability=probability,
                iterations=iterations,
                cld_shad=cld_shad,
                cld_med_prob=cld_med_prob,
                cld_hi_prob=cld_hi_prob,
                thin_cir=thin_cir,
            )

        return self

    def set_permissions(self):
        try:
            set_permissions(self.path)
        except:
            logger.info("Cannot set permissions to: {}".format(self.identifier))

    # def has_indice(self, indice):
    # """
    # Returns if the indice file exists on disk.

    # :param indice: name of the indice.
    # """
    # indice_cls = IndicesCollection().get_indice_cls(indice)

    # if indice_cls is None:
    # raise ValueError("{} is not a valid indice name.".format(indice))

    # template = indice_cls.filename_template
    # ext = indice_cls.ext
    # indice_filename = template.format(product_identifier=self.identifier,
    # ext=ext)
    # return (self.indices_path / indice / self.tile / indice_filename).exists()

    # @property
    # def sen2cor_version(self):
    #     """ Used Sen2Cor version"""
    #     return Sen2ChainMetadataParser(self.identifier):

    #     sen2chain_info_path = self.path / "Sen2Chain_info.xml"
    #     if sen2chain_info_path.exists():
    #         return "xml present"
    #         Sen2ChainMetadataParser(self.identifier)

    #     else:
    #         return None

    def update_md(
        self,
        sen2chain_version: str = None,
        sen2chain_processing_version: str = None,
        sen2cor_version: str = None,
    ):
        """Set custom sen2chain, sen2chain_processing and sen2cor versions"""
        Sen2ChainMetadataParser(self._sen2chain_info_path).set_metadata(
            sen2chain_version=sen2chain_version,
            sen2chain_processing_version=sen2chain_processing_version,
            sen2cor_version=sen2cor_version,
        )

    def remove(self):
        if self.path.is_symlink():
            l2a_path = os.readlink(str(self.path))
            logger.info("Removing: {}".format(l2a_path))
            shutil.rmtree(l2a_path)
            logger.info("Removing symlink: {}".format(self.path))
            self.path.unlink()
        else:
            shutil.rmtree(str(self.path))
            logger.info("Removing: {}".format(self.path))

    @property
    def sen2chain_version(self):
        return Sen2ChainMetadataParser(
            self._sen2chain_info_path
        ).get_metadata_value("SEN2CHAIN_VERSION")

    @property
    def sen2chain_processing_version(self):
        return Sen2ChainMetadataParser(
            self._sen2chain_info_path
        ).get_metadata_value("SEN2CHAIN_PROCESSING_VERSION")

    @property
    def sen2cor_version(self):
        return Sen2ChainMetadataParser(
            self._sen2chain_info_path
        ).get_metadata_value("SEN2COR_VERSION")

    @property
    def generation_time(self):
        return self._get_metadata_value(key="GENERATION_TIME")

    # METADATA
    @property
    def product_uri_1c(self):
        return self._get_metadata_value(key="PRODUCT_URI_1C")

    @property
    def product_uri_2a(self):
        return self._get_metadata_value(key="PRODUCT_URI_2A")

    @property
    def cloud_coverage_assessment(self):
        return self._get_metadata_value(key="Cloud_Coverage_Assessment")

    @property
    def nodata_pixel_percentage(self):
        return self._get_metadata_value(key="NODATA_PIXEL_PERCENTAGE")

    @property
    def saturated_defective_pixel_percentage(self):
        return self._get_metadata_value(
            key="SATURATED_DEFECTIVE_PIXEL_PERCENTAGE"
        )

    @property
    def dark_features_percentage(self):
        return self._get_metadata_value(key="DARK_FEATURES_PERCENTAGE")

    @property
    def cloud_shadow_percentage(self):
        return self._get_metadata_value(key="CLOUD_SHADOW_PERCENTAGE")

    @property
    def vegetation_percentage(self):
        return self._get_metadata_value(key="VEGETATION_PERCENTAGE")

    @property
    def not_vegetated_percentage(self):
        return self._get_metadata_value(key="NOT_VEGETATED_PERCENTAGE")

    @property
    def water_percentage(self):
        return self._get_metadata_value(key="WATER_PERCENTAGE")

    @property
    def unclassified_percentage(self):
        return self._get_metadata_value(key="UNCLASSIFIED_PERCENTAGE")

    @property
    def medium_proba_clouds_percentage(self):
        return self._get_metadata_value(key="MEDIUM_PROBA_CLOUDS_PERCENTAGE")

    @property
    def high_proba_clouds_percentage(self):
        return self._get_metadata_value(key="HIGH_PROBA_CLOUDS_PERCENTAGE")

    @property
    def thin_cirrus_percentage(self):
        return self._get_metadata_value(key="THIN_CIRRUS_PERCENTAGE")

    # FOOTPRINT
    @property
    def footprint(self) -> List[Tuple[float, float]]:
        """Returns product's footprint as list of coordinates couples."""
        fp_str = self._get_metadata_value(key="EXT_POS_LIST")
        fp_list = fp_str.strip().split()
        fp_coords = grouper(fp_list, 2)
        fp_coords_float = [(float(lat), float(lon)) for lon, lat in fp_coords]
        return fp_coords_float

    # BANDS
    # 10m bands
    @property
    def b02_10m(self):
        return self._get_band_path(key="B02", res="10m")

    @property
    def b03_10m(self):
        return self._get_band_path(key="B03", res="10m")

    @property
    def b04_10m(self):
        return self._get_band_path(key="B04", res="10m")

    @property
    def b08_10m(self):
        return self._get_band_path(key="B08", res="10m")

    @property
    def aot_10m(self):
        return self._get_band_path(key="AOT", res="10m")

    @property
    def wvp_10m(self):
        return self._get_band_path(key="WVP", res="10m")

    @property
    def tci_10m(self):
        return self._get_band_path(key="TCI", res="10m")

    # 20m bands
    @property
    def b02_20m(self):
        return self._get_band_path(key="B02", res="20m")

    @property
    def b03_20m(self):
        return self._get_band_path(key="B03", res="20m")

    @property
    def b04_20m(self):
        return self._get_band_path(key="B04", res="20m")

    @property
    def b05_20m(self):
        return self._get_band_path(key="B05", res="20m")

    @property
    def b06_20m(self):
        return self._get_band_path(key="B06", res="20m")

    @property
    def b07_20m(self):
        return self._get_band_path(key="B07", res="20m")

    @property
    def b8a_20m(self):
        return self._get_band_path(key="B8A", res="20m")

    @property
    def b11_20m(self):
        return self._get_band_path(key="B11", res="20m")

    @property
    def b12_20m(self):
        return self._get_band_path(key="B12", res="20m")

    @property
    def scl_20m(self):
        return self._get_band_path(key="SCL", res="20m")

    @property
    def snw_20m(self):
        """old sen2cor snow-mask's name"""
        # try:
        return self._get_band_path(key="SNW")
        # except:
        # logger.warning("'snw_20m deprecated', use: 'msk_snwprb_20m'")
        # return self.msk_snwprb_20m

    @property
    def msk_snwprb_20m(self):
        """new sen2cor cloud-mask's name"""
        # try:
        return self._get_band_path(key="SNW")
        # except:
        # return self._get_band_path(key="SNWPRB", res="20m")

    @property
    def cld_20m(self):
        """old sen2cor cloud-mask's name"""
        # try:
        return self._get_band_path(key="CLD", res="20m")
        # Except:
        # logger.warning("'cld_20m' deprecated', use: 'msk_cldprb_20m'")
        # return self.msk_cldprb_20m

    @property
    def msk_cldprb_20m(self):
        """new sen2cor cloud-mask's name"""
        # try:
        # return self._get_band_path(key="MSK_CLDPRB", res="20m")
        # except:
        return self._get_band_path(key="CLD", res="20m")

    @property
    def aot_20m(self):
        return self._get_band_path(key="AOT", res="20m")

    @property
    def wvp_20m(self):
        return self._get_band_path(key="WVP", res="20m")

    @property
    def tci_20m(self):
        return self._get_band_path(key="TCI", res="20m")

    # 60m bands
    @property
    def b01_60m(self):
        return self._get_band_path(key="B01", res="60m")

    @property
    def b02_60m(self):
        return self._get_band_path(key="B02", res="60m")

    @property
    def b03_60m(self):
        return self._get_band_path(key="B03", res="60m")

    @property
    def b04_60m(self):
        return self._get_band_path(key="B04", res="60m")

    @property
    def b05_60m(self):
        return self._get_band_path(key="B05", res="60m")

    @property
    def b06_60m(self):
        return self._get_band_path(key="B06", res="60m")

    @property
    def b07_60m(self):
        return self._get_band_path(key="B07", res="60m")

    @property
    def b8a_60m(self):
        return self._get_band_path(key="B8A", res="60m")

    @property
    def b09_60m(self):
        return self._get_band_path(key="B09", res="60m")

    @property
    def b11_60m(self):
        return self._get_band_path(key="B11", res="60m")

    @property
    def b12_60m(self):
        return self._get_band_path(key="B12", res="60m")

    @property
    def scl_60m(self):
        return self._get_band_path(key="SCL", res="60m")

    @property
    def snw_60m(self):
        return self._get_band_path(key="SNW", res="60m")

    @property
    def cld_60m(self):
        return self._get_band_path(key="CLD", res="60m")

    @property
    def aot_60m(self):
        return self._get_band_path(key="AOT", res="60m")

    @property
    def wvp_60m(self):
        return self._get_band_path(key="WVP", res="60m")

    @property
    def tci_60m(self):
        return self._get_band_path(key="TCI", res="60m")


class OldCloudMaskProduct:
    """Old cloud mask product class.

    :param identifier: cloudmask filename.

    """

    _library_path = Path(Config().get("l2a_path"))

    def __init__(self, identifier: str = None) -> None:
        if identifier is None:
            raise ValueError("Product identifier is empty")
        else:
            self.identifier = identifier
        self.tile = self.get_tile(identifier)
        self.l2a = self.get_l2a(identifier)
        self.path = self._library_path / self.tile / self.identifier

    @staticmethod
    def get_tile(identifier) -> str:
        """Returns tile name from a string.

        :param string: string from which to extract the tile name.
        """
        return re.findall("_T([0-9]{2}[A-Z]{3})_", identifier)[0]

    @staticmethod
    def get_l2a(identifier) -> str:
        """Returns l2a name from a old cloud mask identifier string.
        :param string: string from which to extract the l2a name.
        """
        return re.findall(r"(S2.+)_CLOUD_MASK.+jp2", identifier)[0]


class NewCloudMaskProduct:
    """New cloud mask product class.
    :param identifier: cloudmask filename.

    """

    _library_path = Path(Config().get("cloudmasks_path"))

    def __init__(
        self,
        identifier: str = None,
        l2a_identifier: str = None,
        sen2chain_processing_version: str = None,
        cm_version: str = "cm001",
        probability: int = 1,
        iterations: int = 5,
        cld_shad: bool = True,
        cld_med_prob: bool = True,
        cld_hi_prob: bool = True,
        thin_cir: bool = True,
    ) -> None:
        if not (identifier or l2a_identifier):
            raise ValueError("Product or L2a identifier cannot be empty")
        else:
            self.tile = self.get_tile(identifier or l2a_identifier)
            self.l2a = (l2a_identifier or self.get_l2a(identifier)).replace(
                ".SAFE", ""
            )
            self.suffix = [
                i
                for i in [
                    "CM001",
                    "CM002-B11",
                    "CM003-PRB" + str(probability) + "-ITER" + str(iterations),
                    "CM004-CSH"
                    + str(int(cld_shad))
                    + "-CMP"
                    + str(int(cld_med_prob))
                    + "-CHP"
                    + str(int(cld_hi_prob))
                    + "-TCI"
                    + str(int(thin_cir))
                    + "-ITER"
                    + str(iterations),
                ]
                if cm_version.upper() in i
            ][0]
            self.identifier = (
                identifier or self.l2a + "_" + self.suffix + ".jp2"
            )
            # self.cm_version, self.probability, self.iterations = self.get_cm_version(self.identifier)
            self.mask_info = self.get_cm_version(self.identifier)

            self.path = (
                self._library_path / self.tile / self.l2a / self.identifier
            )
            self._info_path = self.path.parent / (self.path.stem + ".xml")
            self.init_md()

    @staticmethod
    def get_tile(identifier) -> str:
        """Returns tile name from a string.

        :param string: string from which to extract the tile name.
        """
        return re.findall("_T([0-9]{2}[A-Z]{3})_", identifier)[0]

    @staticmethod
    def get_identifier(l2a_identifier) -> str:
        """Returns l2a name from a old cloud mask identifier string.
        :param string: string from which to extract the l2a name.
        """
        return re.findall(r"(S2.+)_CM.+jp2", identifier)[0]

    # @staticmethod
    # def get_l2a(identifier) -> str:
    # """Returns l2a name from a old cloud mask identifier string.
    # :param string: string from which to extract the l2a name.
    # """
    # return re.findall(r"(S2.+)_CM.+jp2", identifier)[0]

    @staticmethod
    def get_l2a(identifier) -> str:
        """Returns l2a name from a old cloud mask identifier string.
        :param string: string from which to extract the l2a name.
        """
        return re.findall(
            r"(S2._.+_[0-9]{8}T[0-9]{6}_N[0-9]{4}_R[0-9]{3}_T[0-9]{2}[A-Z]{3}_[0-9]{8}T[0-9]{6})_.*",
            identifier,
        )[0]

    @staticmethod
    def get_cm_version(identifier) -> str:
        """Returns cloudmask version from a cloud mask identifier string.
        :param string: string from which to extract the version name.
        """
        try:
            pat = re.compile(r"S2.+_(?P<cm_version>CM00[1-2])")
            return pat.match(identifier).groupdict()
        except:
            pass
        try:
            pat = re.compile(
                r"S2.+_(?P<cm_version>CM003)"
                + "-PRB(?P<probability>.*)"
                + "-ITER(?P<iterations>.*)"
                + ".jp2"
            )
            return pat.match(identifier).groupdict()
        except:
            pass
        try:
            pat = re.compile(
                r"S2.+_(?P<cm_version>CM004)"
                + "-CSH(?P<cld_shad>.*)"
                + "-CMP(?P<cld_med_prob>.*)"
                + "-CHP(?P<cld_hi_prob>.*)"
                + "-TCI(?P<thin_cir>.*)"
                + "-ITER(?P<iterations>.*)"
                + ".jp2"
            )
            return pat.match(identifier).groupdict()
        except:
            pass

        # try:
        #     return re.findall(r"S2.+_(CM004)-CSH(.*)-CMP(.*)-CHP(.*)-TCI(.*)-ITER(.*)\.", identifier)[0]
        # except:
        #     try:
        #         return [re.findall(r"S2.+_(CM003)-PRB(.*)-ITER(.*)\.", identifier)[0], None, None]
        #     except:
        #         try:
        #             return [re.findall(r"S2.+_(CM00[1-2]).+", identifier)[0], None, None]
        #         except:
        #             return [None, None, None]

    @property
    def sen2chain_version(self):
        return Sen2ChainMetadataParser(self._info_path).get_metadata_value(
            "SEN2CHAIN_VERSION"
        )

    @property
    def sen2chain_processing_version(self):
        return Sen2ChainMetadataParser(self._info_path).get_metadata_value(
            "SEN2CHAIN_PROCESSING_VERSION"
        )

    @property
    def sen2cor_version(self):
        return Sen2ChainMetadataParser(self._info_path).get_metadata_value(
            "SEN2COR_VERSION"
        )

    def init_md(self):
        if self.path.exists() and not self._info_path.exists():
            l2a = L2aProduct(self.l2a)
            if l2a._sen2chain_info_path.exists():
                Sen2ChainMetadataParser(self._info_path).set_metadata(
                    sen2chain_version=l2a.sen2chain_version,
                    sen2chain_processing_version=l2a.sen2chain_processing_version,
                    sen2cor_version=l2a.sen2cor_version,
                )
            else:
                Sen2ChainMetadataParser(self._info_path).init_metadata()

    def update_md(
        self,
        sen2chain_version: str = None,
        sen2chain_processing_version: str = None,
        sen2cor_version: str = None,
    ):
        """Set custom sen2chain, sen2chain_processing and sen2cor versions"""
        Sen2ChainMetadataParser(self._info_path).set_metadata(
            sen2chain_version=sen2chain_version,
            sen2chain_processing_version=sen2chain_processing_version,
            sen2cor_version=sen2cor_version,
        )


class IndiceProduct:
    """Indice product class."""

    # _library_path = Path(Config().get("cloudmasks_path"))
    _indices_path = Path(Config().get("indices_path"))

    def __init__(
        self,
        identifier: str = None,
        l2a_identifier: str = None,
        indice: str = None,
        sen2chain_processing_version: str = None,
        masked: bool = False,
        cm_version: str = "cm001",
        probability: int = 1,
        iterations: int = 5,
        cld_shad: bool = True,
        cld_med_prob: bool = True,
        cld_hi_prob: bool = True,
        thin_cir: bool = True,
    ) -> None:
        if not (identifier or (l2a_identifier and indice)):
            raise ValueError(
                "Product or (L2a identifier and indice) cannot be empty"
            )
        else:
            self.tile = NewCloudMaskProduct.get_tile(
                identifier or l2a_identifier
            )
            self.l2a = (
                l2a_identifier or NewCloudMaskProduct.get_l2a(identifier)
            ).replace(".SAFE", "")
            self.indice = (
                indice or identifier.replace(".", "_").split("_")[7]
            ).upper()
            self.masked = masked
            if self.masked:
                # self.suffix = [i for i in ["CM001", "CM002-B11", "CM003-PRB" + str(probability) + "-ITER" + str(iterations)] if cm_version.upper() in i][0]
                self.suffix = [
                    i
                    for i in [
                        "CM001",
                        "CM002-B11",
                        "CM003-PRB"
                        + str(probability)
                        + "-ITER"
                        + str(iterations),
                        "CM004-CSH"
                        + str(int(cld_shad))
                        + "-CMP"
                        + str(int(cld_med_prob))
                        + "-CHP"
                        + str(int(cld_hi_prob))
                        + "-TCI"
                        + str(int(thin_cir))
                        + "-ITER"
                        + str(iterations),
                    ]
                    if cm_version.upper() in i
                ][0]
                self.identifier = (
                    identifier
                    or self.l2a
                    + "_"
                    + self.indice
                    + "_"
                    + self.suffix
                    + ".jp2"
                )
                # self.cm_version, self.probability, self.iterations = NewCloudMaskProduct.get_cm_version(self.identifier)
                self.mask_info = NewCloudMaskProduct.get_cm_version(
                    self.identifier
                )
            else:
                self.suffix = None
                self.identifier = (
                    identifier or self.l2a + "_" + self.indice + ".jp2"
                )
                # self.cm_version, self.probability, self.iterations = 3* [None]
                self.mask_info = None
            self.path = (
                self._indices_path
                / self.indice
                / self.tile
                / self.l2a
                / self.identifier
            )
            # self.cm_version = self.cm_version or cm_version
            # self.probability = self.probability or probability
            # self.iterations = self.iterations or iterations
            self._info_path = self.path.parent / (self.path.stem + ".xml")
            self.init_md()

    @staticmethod
    def get_indice(identifier) -> str:
        """ """
        return re.findall(r"S2.+_(.+)_.*jp2", identifier)[0]

    def init_md(self):
        if self.path.exists() and not self._info_path.exists():
            l2a = L2aProduct(self.l2a)
            if l2a._sen2chain_info_path.exists():
                Sen2ChainMetadataParser(self._info_path).set_metadata(
                    sen2chain_version=l2a.sen2chain_version,
                    sen2chain_processing_version=l2a.sen2chain_processing_version,
                    sen2cor_version=l2a.sen2cor_version,
                )
            else:
                Sen2ChainMetadataParser(self._info_path).init_metadata()
