# coding: utf-8

"""
This module contains usefull functions commons to every modules.
"""

import os, stat
import logging
from itertools import zip_longest
from datetime import datetime
from typing import Iterable, Union
import re
import pathlib
from packaging import version


from .config import Config
# from .products import L1cProduct

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


# useful dates formats encountered when working with Sentinel-2 data
DATES_FORMATS = dict(
    ymd="%Y-%m-%d",
    Ymd = "%Y%m%d",
    filename="%Y%m%dT%H%M%S",
    metadata="%Y-%m-%dT%H:%M:%S.%fZ"
)

def format_word(word: str) -> str:
    """
    Formats a word to lower case and removes heading and trailing spaces.

    :param word: a string to format.
    """
    if not isinstance(word, str):
        raise TypeError("word has to be a string")
    return word.lower().strip()


def grouper(iterable, n: int, fillvalue: bool = None) -> Iterable:
    """
    Collects data into fixed-length chunks or blocks.

    :param iterable: an iterable.
    :param n: number of items per chunk/block.
    :param fillvalue: replacing value for missing items in the chunk.
    """
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


def str_to_datetime(date_string: str, date_format: str) -> datetime:
    """
    Returns a datetime object from a string according to a format.

    :param date_string: string to format.
    :param date_format: date format.
    """
    try:
        date_format = DATES_FORMATS.get(date_format)
    except KeyError:
        raise ValueError("Invalid date format")
    return datetime.strptime(date_string, date_format)


def datetime_to_str(date: datetime, date_format: str) -> str:
    """
    Returns a string from a datetime object according to a format.

    :param date: string to format.
    :param date_format: date format.
    """
    try:
        date_format = DATES_FORMATS.get(date_format)
    except KeyError:
        raise ValueError("Invalid date format")
    return datetime.strftime(date, date_format)


def human_size(bytes, units=["B", "KB", "MB", "GB", "TB", "PB", "EB"]):
    """Returns a human readable string reprentation of bytes"""
    return (
        str(bytes) + units[0]
        if bytes < 1024.0
        else human_size(bytes >> 10, units[1:])
    )


def human_size_decimal(size, decimal_places=2):
    for unit in ["B", "KB", "MB", "GB", "TB", "PB", "EB"]:
        if size < 1024.0:
            break
        size /= 1024.0
    return "{}{}".format(format(size, "." + str(decimal_places) + "f"), unit)
    # return f"{size:.{decimal_places}f}{unit}"


def getFolderSize(folder, follow_symlinks=False):
    total_size = os.path.getsize(folder)
    try:
        for item in os.listdir(folder):
            itempath = os.path.join(folder, item)
            if os.path.isfile(itempath) and (
                follow_symlinks or not os.path.islink(itempath)
            ):
                total_size += os.path.getsize(itempath)
            elif os.path.isdir(itempath) and (
                follow_symlinks or not os.path.islink(itempath)
            ):
                total_size += getFolderSize(itempath)
    except:
        pass
    return total_size

def set_permissions(path):
    os.chmod(
        str(path), os.stat(str(path)).st_mode | stat.S_IWGRP | stat.S_IWUSR
    )
    os.chmod(str(path), os.stat(str(path)).st_mode & ~stat.S_IWOTH)
    for dir_path, dir_names, files in os.walk(str(path)):
        for d in dir_names:
            name = dir_path + "/" + d
            # os.chmod(name, os.stat(name).st_mode | 0o075)
            # os.chmod(name, os.stat(name).st_mode & ~0o002)
            os.chmod(name, os.stat(name).st_mode | stat.S_IWGRP | stat.S_IWUSR)
            os.chmod(name, os.stat(name).st_mode & ~stat.S_IWOTH)

        for f in files:
            name = dir_path + "/" + f
            # os.chmod(name, os.stat(name).st_mode | 0o064)
            # os.chmod(name, os.stat(name).st_mode & ~0o003)
            os.chmod(
                name, 
                os.stat(name).st_mode | stat.S_IWGRP | stat.S_IWUSR
            )
            os.chmod(
                name, 
                os.stat(name).st_mode & ~stat.S_IXOTH & ~stat.S_IWOTH
            )

def get_current_Sen2Cor_version():
    """Returns your current Sen2Cor version"""
    sen2cor_bashrc_path = Config().get("sen2cor_bashrc_path")
    return next(
        iter(
            re.findall(
                "Sen2Cor-(\d{2}\.\d{2}\.\d{2})", str(sen2cor_bashrc_path)
            )
        ),
        None,
    )

def get_tile(identifier) -> str:
    """Returns tile name from a string.

    :param string: string from which to extract the tile name.
    """
    return re.findall("_T([0-9]{2}[A-Z]{3})_", identifier)[0]
            
def get_Sen2Cor_version(path: Union[str, pathlib.PosixPath] = None):
    """ Returns 
        - the current Sen2Cor version if no argument provided
        - the Sen2Cor version if a Sen2Cor path is provided
    """
    if not path:
        path = Config().get("sen2cor_bashrc_path")
    return next(iter(re.findall('Sen2Cor-(\d{2}\.\d{2}\.\d{2})', str(path))), None)

def get_latest_s2c_version_path(l1c_identifier):
    """ Returns 
    - the path of the first Sen2Cor compatible version, starting from sen2cor_bashrc_path then sen2cor_alternative_bashrc_path
    - None if no compatible Sen2Cor version is provided
    """   
    from .products import L1cProduct
    current_path = Config().get("sen2cor_bashrc_path")
    alternate_path = Config().get("sen2cor_alternative_bashrc_path")
    
    if version.parse(L1cProduct(l1c_identifier).processing_baseline) >= version.parse("4.0"):
        if version.parse(get_Sen2Cor_version(current_path)) >= version.parse("2.10.01"):
            return current_path
        else:
            if version.parse(get_Sen2Cor_version(alternate_path)) >= version.parse("2.10.01"):
                return alternate_path
            else:
                return None
    elif version.parse(L1cProduct(l1c_identifier).processing_baseline) >= version.parse("2.06"):
        return current_path
    else:
        if version.parse(get_Sen2Cor_version(current_path)) >= version.parse("2.10.01"):
            if alternate_path:
                if version.parse(get_Sen2Cor_version(alternate_path)) >= version.parse("2.10.01"):
                    return None
                else:
                    return alternate_path
            else:
                return None
        else:
            return current_path

